#![feature(error_in_core)]
#![cfg_attr(not(any(feature = "std", test)), no_std)]

use core::{borrow::Borrow, mem::size_of, ops::RangeBounds};

use alloc::{
    collections::{btree_map::Entry, BTreeMap},
    vec,
    vec::Vec,
};
use scroll::{Endian, Pread};
use thiserror::Error;

use structs::{BTreeNode, Chunk, Header, Item, Key, KeyPtr, ObjectType, Stripe, Superblock};

extern crate alloc;

pub mod structs;

#[cfg(feature = "std")]
pub mod std_io {
    use std::io::{Cursor, Read, Seek};

    use crate::{Error, VolumeIo};

    impl<T: Read + Seek> VolumeIo for T {
        fn read(&mut self, dst: &mut [u8], address: u64) -> Result<(), Error> {
            self.seek(std::io::SeekFrom::Start(address))
                .map_err(|a| Error::SeekFailed)?;
            self.read_exact(dst).map_err(|_| Error::ReadFailed)
        }

        fn alignment(&self) -> usize {
            1
        }
    }
}

#[derive(Debug, Error)]
pub enum Error {
    #[error("read failed")]
    ReadFailed,
    #[error("seek failed")]
    SeekFailed,
    #[error("invalid magic signature")]
    InvalidMagic,
    #[error("invalid offset")]
    InvalidOffset,
    #[error("Expected an internal node")]
    ExpectedInternalNode,
    #[error("Expected a leaf node")]
    ExpectedLeafNode,
    #[error("{0}")]
    ScrollError(scroll::Error),
}

impl From<scroll::Error> for Error {
    fn from(value: scroll::Error) -> Self {
        Self::ScrollError(value)
    }
}

pub type Result<T> = core::result::Result<T, Error>;

pub trait VolumeIo {
    fn read(&mut self, dst: &mut [u8], address: u64) -> Result<()>;
    fn alignment(&self) -> usize;
}

/// equal if overlapping, ordered by lower bound
#[derive(Debug)]
pub struct ChunkTreeKey {
    range: core::ops::Range<u64>,
}

impl From<u64> for ChunkTreeKey {
    fn from(value: u64) -> Self {
        Self {
            range: value..value,
        }
    }
}

impl ChunkTreeKey {
    pub fn start(&self) -> u64 {
        self.range.start
    }
    pub fn end(&self) -> u64 {
        self.range.end
    }
    pub fn size(&self) -> u64 {
        self.range.end - self.range.start
    }

    pub fn delta(&self, point: u64) -> u64 {
        point - self.range.start
    }

    pub fn sub_range(&self, point: u64) -> core::ops::Range<u64> {
        self.delta(point)..self.end()
    }
}

impl Eq for ChunkTreeKey {}
impl Ord for ChunkTreeKey {
    fn cmp(&self, other: &Self) -> core::cmp::Ordering {
        self.partial_cmp(other).unwrap()
    }
}

impl PartialEq for ChunkTreeKey {
    fn eq(&self, other: &Self) -> bool {
        (self.range.contains(&other.range.start) || self.range.contains(&other.range.end))
            || (other.range.contains(&self.range.start) || other.range.contains(&self.range.end))
    }
}

impl PartialOrd for ChunkTreeKey {
    fn partial_cmp(&self, other: &Self) -> Option<core::cmp::Ordering> {
        self.eq(other)
            .then_some(core::cmp::Ordering::Equal)
            .or_else(|| self.range.start.partial_cmp(&other.range.start))
    }
}

pub struct ChunkCacheTree {
    inner: BTreeMap<ChunkTreeKey, u64>,
}

type ChunkTree = BTreeMap<ChunkTreeKey, u64>;

pub struct Volume<R: VolumeIo> {
    reader: R,
    superblock: Superblock,
    chunk_cache: ChunkTree,
}

impl<R: VolumeIo> Volume<R> {
    pub fn new(mut reader: R) -> Result<Self> {
        let mut buf = vec![0; size_of::<Superblock>()];
        reader.read(&mut buf, Superblock::SUPERBLOCK_BASE_OFFSET as _)?;
        let superblock = Superblock::parse(&buf)?;
        let chunk_cache = Self::bootstrap_chunk_tree(&superblock)?;
        Ok(Self {
            reader,
            superblock,
            chunk_cache,
        })
    }

    pub fn size_from_logical(&self, logical: u64) -> Option<u64> {
        self.chunk_cache
            .get_key_value(&logical.into())
            .map(|(key, _)| key.size())
    }

    pub fn offset_from_logical(&self, logical: u64) -> Option<u64> {
        self.chunk_cache
            .get_key_value(&logical.into())
            .map(|(key, offset)| offset + key.delta(logical))
    }

    pub fn range_from_logical(&self, logical: u64) -> Option<core::ops::Range<u64>> {
        self.chunk_cache
            .get_key_value(&logical.into())
            .map(|(key, offset)| {
                let delta = key.delta(logical);
                (offset + delta)..(offset + key.size() - delta)
            })
    }

    pub fn read_range(&mut self, range: core::ops::Range<u64>) -> Result<Vec<u8>> {
        let mut buf = vec![0; (range.end - range.start) as usize];
        self.reader.read(&mut buf, range.start)?;

        Ok(buf)
    }

    pub fn asdf(&mut self) -> Result<()> {
        let chunk_root = self.superblock.chunk_root;
        // let size = self.size_from_logical(chunk_root).expect("size");

        log::debug!("chunk_root: {chunk_root}");
        let contains = self.chunk_cache.contains_key(&chunk_root.into());
        log::debug!("chunk_cache: {contains} {:?}", self.chunk_cache);
        let physical = self.range_from_logical(chunk_root).expect("range");

        let root = self.read_range(physical)?;

        let node = BTreeNode::parse(&root)?;

        log::debug!("{node:#?}");

        Ok(())
    }

    pub fn bootstrap_chunk_tree(superblock: &Superblock) -> Result<ChunkTree> {
        let array_size = superblock.sys_chunk_array_size as usize;
        let mut offset: usize = 0;

        let key_size = size_of::<Key>();
        let mut chunk_tree = ChunkTree::new();

        let bytes = &superblock.sys_chunk_array;

        while offset < array_size {
            if offset + key_size > array_size {
                log::error!("short key read");
                return Err(Error::InvalidOffset);
            }

            let key = bytes.gread_with::<Key>(&mut offset, scroll::LE)?;
            if key.ty.as_type() != Some(ObjectType::ChunkItem) {
                log::error!("key is not of type ChunkItem");
                return Err(Error::InvalidOffset);
            }

            let chunk = bytes.gread_with::<Chunk>(&mut offset, scroll::LE)?;
            if chunk.num_stripes == 0 {
                log::error!("num_stripes cannot be 0");
                return Err(Error::InvalidOffset);
            }

            let num_stripes = chunk.num_stripes; // copy to prevent unaligned access
            if num_stripes != 1 {
                log::warn!(
                    "warning: {} stripes detected but only processing 1",
                    num_stripes
                );
            }

            match chunk_tree.entry(ChunkTreeKey {
                range: key.offset..(key.offset + chunk.length),
            }) {
                Entry::Vacant(entry) => {
                    entry.insert(chunk.stripe.offset);
                }
                Entry::Occupied(_) => {
                    log::error!("overlapping stripes!");
                    return Err(Error::InvalidOffset);
                }
            };

            offset += (num_stripes - 1) as usize * size_of::<Stripe>();
            if offset > array_size {
                log::error!("short chunk item + stripes read");
                return Err(Error::InvalidOffset);
            }
        }

        Ok(chunk_tree)
    }

    pub fn superblock(&self) -> &Superblock {
        &self.superblock
    }
}
