use core::{mem::size_of, ops::Deref};

use bytemuck::{Pod, Zeroable};
use derivative::Derivative;
use num_enum::{FromPrimitive, TryFromPrimitive};
use scroll::{
    ctx::{SizeWith, TryFromCtx},
    Pread, SizeWith,
};
use zerocopy::{AsBytes, FromBytes};

use crate::{Error, Result};

#[repr(u8)]
#[non_exhaustive]
#[derive(Debug, Clone, Copy, PartialEq, Eq, TryFromPrimitive)]
//#[rustc_nonnull_optimization_guaranteed]
pub enum ObjectType {
    INodeItem = 0x01,
    InodeRef = 0x0C,
    InodeExtref = 0x0D,
    XattrItem = 0x18,
    OrphanInode = 0x30,
    DirItem = 0x54,
    DirIndex = 0x60,
    ExtentData = 0x6C,
    ExtentCsum = 0x80,
    RootItem = 0x84,
    TypeRootBackref = 0x90,
    RootRef = 0x9C,
    ExtentItem = 0xA8,
    MetadataItem = 0xA9,
    TreeBlockRef = 0xB0,
    ExtentDataRef = 0xB2,
    ExtentRefV0 = 0xB4,
    SharedBlockRef = 0xB6,
    SharedDataRef = 0xB8,
    BlockGroupItem = 0xC0,
    FreeSpaceInfo = 0xC6,
    FreeSpaceExtent = 0xC7,
    FreeSpaceBitmap = 0xC8,
    DevExtent = 0xCC,
    DevItem = 0xD8,
    ChunkItem = 0xE4,
    TempItem = 0xF8,
    DevStats = 0xF9,
    SubvolUuid = 0xFB,
    SubvolRecUuid = 0xFC,
}

#[repr(C, packed)]
#[derive(Clone, Copy, PartialEq, Eq, Pread, SizeWith)]
pub struct ObjectTypeWrapper {
    inner: u8,
}

impl core::fmt::Debug for ObjectTypeWrapper {
    fn fmt(&self, f: &mut core::fmt::Formatter<'_>) -> core::fmt::Result {
        f.debug_struct("ObjectTypeWrapper")
            .field("inner", &self.as_type())
            .finish()
    }
}

impl ObjectTypeWrapper {
    pub fn as_type(self) -> Option<ObjectType> {
        ObjectType::try_from_primitive(self.inner).ok()
    }
}

unsafe impl Pod for ObjectTypeWrapper {}
unsafe impl Zeroable for ObjectTypeWrapper {}

#[repr(transparent)]
#[derive(Debug, Clone, Copy)]
pub struct Uuid(uuid::Uuid);

impl SizeWith<scroll::Endian> for Uuid {
    fn size_with(_: &scroll::Endian) -> usize {
        size_of::<Self>()
    }
}

impl<'a> TryFromCtx<'a, scroll::Endian> for Uuid {
    type Error = scroll::Error;

    fn try_from_ctx(
        from: &'a [u8],
        _: scroll::Endian,
    ) -> core::result::Result<(Self, usize), Self::Error> {
        let size = size_of::<Self>();
        Ok((*bytemuck::from_bytes::<Self>(&from[..size]), size))
    }
}

impl Deref for Uuid {
    type Target = uuid::Uuid;

    fn deref(&self) -> &Self::Target {
        &self.0
    }
}

unsafe impl Pod for Uuid {}
unsafe impl Zeroable for Uuid {}

#[repr(C, packed(1))]
#[derive(Debug, Clone, Copy, Pod, Zeroable, Pread, SizeWith)]
pub struct Key {
    pub id: u64,
    pub ty: ObjectTypeWrapper,
    pub offset: u64,
}

#[repr(C, packed(1))]
#[derive(Debug, Clone, Copy, Pod, Zeroable)]
pub struct TreeHeader {
    csum: [u8; 32],
    fs_uuid: Uuid,
    address: u64,
    flags: u64,
    chunk_tree_uuid: Uuid,
    generation: u64,
    tree_id: u64,
    num_items: u32,
    level: u8,
}

#[repr(C, packed(1))]
#[derive(Debug, Clone, Copy, Pod, Zeroable)]
pub struct LeafNode {
    key: Key,
    offset: u32,
    size: u32,
}

#[repr(C, packed(1))]
#[derive(Debug, Clone, Copy, Pod, Zeroable)]
pub struct InternalNode {
    key: Key,
    address: u64,
    generation: u64,
}

const MAX_LABEL_SIZE: usize = 0x100;
const SYS_CHUNK_ARRAY_SIZE: usize = 0x800;
const BTRFS_NUM_BACKUP_ROOTS: usize = 4;

fn format_u8str<T: AsRef<[u8]>>(s: &T, f: &mut core::fmt::Formatter) -> core::fmt::Result {
    let bytes = s.as_ref();
    let end = bytes
        .iter()
        .position(|&b| b == 0)
        .map(|i| i + 1)
        .unwrap_or(bytes.len());
    core::ffi::CStr::from_bytes_with_nul(&bytes[..end])
        .map(|s| write!(f, "{:?}", s))
        .map_err(|_| core::fmt::Error)?
}

#[repr(C, packed(1))]
#[derive(Derivative, Clone, Copy, Pod, Zeroable)]
#[derivative(Debug)]
pub struct INodeItem {
    generation: u64,
    transid: u64,
    st_size: u64,
    st_blocks: u64,
    block_group: u64,
    st_nlink: u32,
    st_uid: u32,
    st_gid: u32,
    st_mode: u32,
    st_rdev: u64,
    flags: u64,
    sequence: u64,
    #[derivative(Debug = "ignore")]
    reserved: [u8; 32],
    st_atime: Timespec,
    st_ctime: Timespec,
    st_mtime: Timespec,
    otime: Timespec,
}

#[repr(C, packed(1))]
#[derive(Debug, Clone, Copy, Pod, Zeroable)]
pub struct ChunkItem {
    size: u64,
    root_id: u64,
    stripe_length: u64,
    ty: u64,
    opt_io_alignment: u32,
    opt_io_width: u32,
    sector_size: u32,
    num_stripes: u16,
    sub_stripes: u16,
}

#[repr(C, packed(1))]
#[derive(Debug, Clone, Copy, Pod, Zeroable)]
pub struct ChunkItemStripe {
    dev_id: u64,
    offset: u64,
    dev_uuid: Uuid,
}

#[repr(C, packed(1))]
#[derive(Debug, Clone, Copy, Pod, Zeroable)]
pub struct ExtentData {
    generation: u64,
    decoded_size: u64,
    compression: u8,
    encryption: u8,
    encoding: u16,
    ty: u8,
    data: [u8; 1],
}

#[repr(C, packed(1))]
#[derive(Debug, Clone, Copy, Pod, Zeroable)]
pub struct ExtentData2 {
    address: u64,
    size: u64,
    offset: u64,
    num_bytes: u64,
}

#[repr(C, packed(1))]
#[derive(Debug, Clone, Copy, Pod, Zeroable)]
pub struct INodeRef {
    index: u64,
    n: u16,
    name: [u8; 1],
}

#[repr(C, packed(1))]
#[derive(Debug, Clone, Copy, Pod, Zeroable)]
pub struct INodeExtRef {
    dir: u64,
    index: u64,
    n: u16,
    name: [u8; 1],
}

#[repr(C, packed(1))]
#[derive(Debug, Clone, Copy, Pod, Zeroable)]
pub struct ExtentItem {
    ref_count: u64,
    generation: u64,
    flags: u64,
}

#[repr(C, packed(1))]
#[derive(Debug, Clone, Copy, Pod, Zeroable)]
pub struct ExtentItem2 {
    first_item: Key,
    level: u8,
}

#[repr(C, packed(1))]
#[derive(Debug, Clone, Copy, Pod, Zeroable)]
pub struct ExtentItemV0 {
    ref_count: u32,
}

#[repr(C, packed(1))]
#[derive(Debug, Clone, Copy, Pod, Zeroable)]
pub struct ExtentItemTree {
    extent_item: ExtentItem,
    first_item: Key,
    level: u8,
}

#[repr(C, packed(1))]
#[derive(Debug, Clone, Copy, Pod, Zeroable)]
pub struct TreeBlockRef {
    offset: u64,
}

#[repr(C, packed(1))]
#[derive(Debug, Clone, Copy, Pod, Zeroable)]
pub struct ExtentDataRef {
    root: u64,
    objid: u64,
    offset: u64,
    count: u32,
}

#[repr(C, packed(1))]
#[derive(Debug, Clone, Copy, Pod, Zeroable)]
pub struct BlockGroupItem {
    used: u64,
    chunk_tree: u64,
    flags: u64,
}

#[repr(C, packed(1))]
#[derive(Debug, Clone, Copy, Pod, Zeroable)]
pub struct ExtentRefV0 {
    root: u64,
    gen: u64,
    objid: u64,
    count: u32,
}

#[repr(C, packed(1))]
#[derive(Debug, Clone, Copy, Pod, Zeroable)]
pub struct SharedBlockRef {
    offset: u64,
}

#[repr(C, packed(1))]
#[derive(Debug, Clone, Copy, Pod, Zeroable)]
pub struct SharedDataRef {
    offset: u64,
    count: u32,
}

#[repr(C, packed(1))]
#[derive(Debug, Clone, Copy, Pod, Zeroable)]
pub struct FreeSpaceEntry {
    offset: u64,
    size: u64,
    ty: u8,
}

#[repr(C, packed(1))]
#[derive(Debug, Clone, Copy, Pod, Zeroable)]
pub struct FreeSpaceItem {
    key: Key,

    generation: u64,
    num_entries: u64,
    num_bitmaps: u64,
}

#[repr(C, packed(1))]
#[derive(Debug, Clone, Copy, Pod, Zeroable)]
pub struct RootRef {
    dir: u64,
    index: u64,
    n: u16,
    name: [u8; 1],
}

#[repr(C, packed(1))]
#[derive(Debug, Clone, Copy, Pod, Zeroable)]
pub struct DevExtent {
    chunktree: u64,
    objid: u64,
    address: u64,
    length: u64,
    chunktree_uuid: Uuid,
}

#[repr(C, packed(1))]
#[derive(Debug, Clone, Copy, Pod, Zeroable)]
pub struct BalanceArgs {}

#[repr(C, packed(1))]
#[derive(Debug, Clone, Copy, Pod, Zeroable)]
pub struct BalanceItem {}

#[repr(C, packed(1))]
#[derive(Debug, Clone, Copy, Pod, Zeroable)]
pub struct FreeSpaceInfo {
    count: u32,
    flags: u32,
}

#[repr(C, packed(1))]
#[derive(Debug, Clone, Copy, Pod, Zeroable)]
pub struct SendHeader {}
#[repr(C, packed(1))]
#[derive(Debug, Clone, Copy, Pod, Zeroable)]
pub struct SendCommand {}
#[repr(C, packed(1))]
#[derive(Debug, Clone, Copy, Pod, Zeroable)]
pub struct SendTlv {}

const CSUM_SIZE: usize = 32;
const LABEL_SIZE: usize = 256;
const SYSTEM_CHUNK_ARRAY_SIZE: usize = 2048;

pub const FS_TREE_OBJECTID: u64 = 5;

pub const INODE_REF_KEY: u8 = 12;
pub const DIR_ITEM_KEY: u8 = 84;
pub const ROOT_ITEM_KEY: u8 = 132;
pub const CHUNK_ITEM_KEY: u8 = 228;

pub const FT_REG_FILE: u8 = 1;

#[repr(C, packed)]
#[derive(Debug, Clone, Copy, Pod, Zeroable, Pread, SizeWith)]
pub struct DevItem {
    /// the internal btrfs device id
    pub devid: u64,
    /// size of the device
    pub total_bytes: u64,
    /// bytes used
    pub bytes_used: u64,
    /// optimal io alignment for this device
    pub io_align: u32,
    /// optimal io width for this device
    pub io_width: u32,
    /// minimal io size for this device
    pub sector_size: u32,
    /// type and info about this device
    pub ty: u64,
    /// expected generation for this device
    pub generation: u64,
    /// starting byte of this partition on the device, to allow for stripe alignment in the future
    pub start_offset: u64,
    /// grouping information for allocation decisions
    pub dev_group: u32,
    /// seek speed 0-100 where 100 is fastest
    pub seek_speed: u8,
    /// bandwidth 0-100 where 100 is fastest
    pub bandwidth: u8,
    /// btrfs generated uuid for this device
    pub uuid: Uuid,
    /// uuid of FS who owns this device
    pub fsid: Uuid,
}

#[repr(C, packed)]
#[derive(Debug, Clone, Copy, Pod, Zeroable, Pread, SizeWith)]
pub struct RootBackup {
    pub tree_root: u64,
    pub tree_root_gen: u64,
    pub chunk_root: u64,
    pub chunk_root_gen: u64,
    pub extent_root: u64,
    pub extent_root_gen: u64,
    pub fs_root: u64,
    pub fs_root_gen: u64,
    pub dev_root: u64,
    pub dev_root_gen: u64,
    pub csum_root: u64,
    pub csum_root_gen: u64,
    pub total_bytes: u64,
    pub bytes_used: u64,
    pub num_devices: u64,
    /// future
    pub unused_64: [u64; 4],
    pub tree_root_level: u8,
    pub chunk_root_level: u8,
    pub extent_root_level: u8,
    pub fs_root_level: u8,
    pub dev_root_level: u8,
    pub csum_root_level: u8,
    /// future and to align
    pub unused_8: [u8; 10],
}

#[repr(C, packed)]
#[derive(Derivative, Clone, Copy, Pod, Zeroable, Pread, SizeWith)]
#[derivative(Debug)]
pub struct Superblock {
    pub csum: [u8; 32],
    pub fsid: Uuid,
    /// Physical address of this block
    pub bytenr: u64,
    pub flags: u64,
    pub magic: [u8; 0x8],
    pub generation: u64,
    /// Logical address of the root tree root
    pub root: u64,
    /// Logical address of the chunk tree root
    pub chunk_root: u64,
    /// Logical address of the log tree root
    pub log_root: u64,
    pub log_root_transid: u64,
    pub total_bytes: u64,
    pub bytes_used: u64,
    pub root_dir_objectid: u64,
    pub num_devices: u64,
    pub sector_size: u32,
    pub node_size: u32,
    /// Unused and must be equal to `nodesize`
    pub leafsize: u32,
    pub stripesize: u32,
    pub sys_chunk_array_size: u32,
    pub chunk_root_generation: u64,
    pub compat_flags: u64,
    pub compat_ro_flags: u64,
    pub incompat_flags: u64,
    pub csum_type: u16,
    pub root_level: u8,
    pub chunk_root_level: u8,
    pub log_root_level: u8,
    pub dev_item: DevItem,
    #[derivative(Debug(format_with = "format_u8str"))]
    pub label: [u8; 0x100],
    pub cache_generation: u64,
    pub uuid_tree_generation: u64,
    pub metadata_uuid: Uuid,
    /// Future expansion
    reserved: [u64; 28],
    #[derivative(Debug = "ignore")]
    pub sys_chunk_array: [u8; 0x800],
    pub root_backup0: RootBackup,
    pub root_backup1: RootBackup,
    pub root_backup2: RootBackup,
    pub root_backup3: RootBackup,
}

#[repr(u16)]
#[derive(Derivative, Clone, Copy, TryFromPrimitive)]
#[derivative(Debug)]
pub enum ChecksumType {
    Crc32 = 0,
    XxHash64,
    Sha256,
    Blake2B,
}

impl Superblock {
    pub const SUPERBLOCK_BASE_OFFSET: usize = 0x10000;
    pub const SUPERBLOCK_OFFSETS: [usize; 4] = [
        Self::SUPERBLOCK_BASE_OFFSET,
        0x4000000,
        0x4000000000,
        0x4000000000000,
    ];
    pub const MAGIC: [u8; 8] = *b"_BHRfS_M";

    pub fn parse(bytes: &[u8]) -> Result<Self> {
        let superblock = bytes.pread_with::<Self>(0, scroll::LE)?;

        if !superblock.verify_magic() {
            return Err(Error::InvalidMagic);
        }

        Ok(superblock)
    }

    pub fn checksum_type(&self) -> Option<ChecksumType> {
        ChecksumType::try_from_primitive(self.csum_type).ok()
    }

    pub fn verify_magic(&self) -> bool {
        self.magic == Self::MAGIC
    }
}

#[repr(C, packed)]
#[derive(Debug, Clone, Copy, Pod, Zeroable, Pread)]
pub struct Stripe {
    pub devid: u64,
    pub offset: u64,
    pub dev_uuid: Uuid,
}

impl Stripe {
    pub fn parse(bytes: &[u8]) -> Result<Self> {
        Ok(bytes.pread_with(0, scroll::LE)?)
    }
}

#[repr(C, packed)]
#[derive(Debug, Clone, Copy, Pod, Zeroable, Pread)]
pub struct Chunk {
    /// size of this chunk in bytes
    pub length: u64,
    /// objectid of the root referencing this chunk
    pub owner: u64,
    pub stripe_len: u64,
    pub ty: u64,
    /// optimal io alignment for this chunk
    pub io_align: u32,
    /// optimal io width for this chunk
    pub io_width: u32,
    /// minimal io size for this chunk
    pub sector_size: u32,
    /// 2^16 stripes is quite a lot, a second limit is the size of a single item in the btree
    pub num_stripes: u16,
    /// sub stripes only matter for raid10
    pub sub_stripes: u16,
    pub stripe: Stripe,
    // additional stripes go here
}

impl Chunk {
    pub fn parse(bytes: &[u8]) -> Result<Self> {
        Ok(bytes.pread_with(0, scroll::LE)?)
    }
}

#[repr(C, packed)]
#[derive(Debug, Clone, Copy, Pod, Zeroable, Pread)]
pub struct Timespec {
    pub sec: u64,
    pub nsec: u32,
}

#[repr(C, packed)]
#[derive(Debug, Clone, Copy, Pod, Zeroable)]
pub struct InodeItem {
    /// nfs style generation number
    pub generation: u64,
    /// transid that last touched this inode
    pub transid: u64,
    pub size: u64,
    pub nbytes: u64,
    pub block_group: u64,
    pub nlink: u32,
    pub uid: u32,
    pub gid: u32,
    pub mode: u32,
    pub rdev: u64,
    pub flags: u64,
    /// modification sequence number for NFS
    pub sequence: u64,
    pub reserved: [u64; 4],
    pub atime: Timespec,
    pub ctime: Timespec,
    pub mtime: Timespec,
    pub otime: Timespec,
}

#[repr(C, packed)]
#[derive(Debug, Clone, Copy, Pod, Zeroable)]
pub struct RootItem {
    pub inode: InodeItem,
    pub generation: u64,
    pub root_dirid: u64,
    pub bytenr: u64,
    pub byte_limit: u64,
    pub bytes_used: u64,
    pub last_snapshot: u64,
    pub flags: u64,
    pub refs: u32,
    pub drop_progress: Key,
    pub drop_level: u8,
    pub level: u8,
    pub generation_v2: u64,
    pub uuid: Uuid,
    pub parent_uuid: Uuid,
    pub received_uuid: Uuid,
    /// updated when an inode changes
    pub ctransid: u64,
    /// trans when created
    pub otransid: u64,
    /// trans when sent. non-zero for received subvol
    pub stransid: u64,
    /// trans when received. non-zero for received subvol
    pub rtransid: u64,
    pub ctime: Timespec,
    pub otime: Timespec,
    pub stime: Timespec,
    pub rtime: Timespec,
    pub reserved: [u64; 8],
}

#[repr(C, packed)]
#[derive(Debug, Clone, Copy, Pod, Zeroable)]
pub struct DirItem {
    pub location: Key,
    pub transid: u64,
    pub data_len: u16,
    pub name_len: u16,
    pub ty: u8,
}

#[repr(C, packed)]
#[derive(Debug, Clone, Copy, Pod, Zeroable)]
pub struct InodeRef {
    pub index: u64,
    pub name_len: u16,
}

#[repr(C, packed)]
#[derive(Debug, Clone, Copy, Pod, Zeroable, Pread)]
pub struct Header {
    pub csum: [u8; 32],
    pub fsid: Uuid,
    /// Which block this node is supposed to live in
    pub bytenr: u64,
    pub flags: u64,
    pub chunk_tree_uuid: Uuid,
    pub generation: u64,
    pub owner: u64,
    pub nritems: u32,
    pub level: u8,
}

#[repr(C, packed)]
#[derive(Debug, Clone, Copy, Pod, Zeroable, Pread)]
/// A `BtrfsLeaf` is full of `BtrfsItem`s. `offset` and `size` (relative to start of data area)
/// tell us where to find the item in the leaf.
pub struct Item {
    pub key: Key,
    pub offset: u32,
    pub size: u32,
}

#[repr(C, packed)]
#[derive(Debug, Clone, Copy, Pod, Zeroable)]
pub struct Leaf {
    pub header: Header,
    // `Item`s begin here
}

#[repr(C, packed)]
#[derive(Debug, Clone, Copy, Pod, Zeroable, Pread)]
/// All non-leaf blocks are nodes and they hold only keys are pointers to other blocks
pub struct KeyPtr {
    pub key: Key,
    pub blockptr: u64,
    pub generation: u64,
}

#[repr(C, packed)]
#[derive(Debug, Clone, Copy, Pod, Zeroable)]
pub struct Node {
    pub header: Header,
    // `KeyPtr`s begin here
}

use alloc::vec::Vec;

#[derive(Debug)]
pub struct BTreeLeafNode {
    header: Header,
    /// actual leaf data
    items: Vec<Item>,
}

impl BTreeLeafNode {
    pub fn parse(header: Header, bytes: &[u8]) -> Result<Self> {
        let offset = &mut 0;
        let items = core::iter::from_fn(|| {
            if *offset as usize + size_of::<Item>() < bytes.len() {
                Some(bytes.gread_with::<Item>(offset, scroll::LE))
            } else {
                None
            }
        })
        .take(header.nritems as usize)
        .collect::<core::result::Result<Vec<_>, _>>()?;
        Ok(Self { header, items })
    }
}

#[derive(Debug)]
pub struct BTreeInternalNode {
    header: Header,
    children: Vec<KeyPtr>,
}

impl BTreeInternalNode {
    pub fn parse(header: Header, bytes: &[u8]) -> Result<Self> {
        let offset = &mut 0;
        let children = core::iter::from_fn(|| {
            if *offset as usize + size_of::<KeyPtr>() < bytes.len() {
                Some(bytes.gread_with::<KeyPtr>(offset, scroll::LE))
            } else {
                None
            }
        })
        .take(header.nritems as usize)
        .collect::<core::result::Result<Vec<_>, _>>()?;
        Ok(Self { header, children })
    }
}

#[derive(Debug)]
pub enum BTreeNode {
    Internal(BTreeInternalNode),
    Leaf(BTreeLeafNode),
}

impl BTreeNode {
    pub fn parse(bytes: &[u8]) -> Result<Self> {
        let offset = &mut 0;
        let header = bytes.gread_with::<Header>(offset, scroll::LE)?;

        if header.level == 0 {
            Ok(Self::Leaf(BTreeLeafNode::parse(header, &bytes[*offset..])?))
        } else {
            Ok(Self::Internal(BTreeInternalNode::parse(
                header,
                &bytes[*offset..],
            )?))
        }
    }
}
